/*************************************************************************
      > File Name:
      > Author:
      > Mail:
      > Created Time:
 ************************************************************************/

#include <fstream>
#include <iomanip>
#include <iostream>
#include <string>
#include <unistd.h>
#include <vector>
#include <cmath>
#include <numeric>
#include <cassert>
#include <stdexcept>
#include <cuda_runtime.h>
#include <cuda.h>
#include "transpose.h"

using namespace std;
#define ck(call) check(call, __LINE__, __FILE__)

inline bool check(cudaError_t e, int iLine, const char *szFile)
{
    if (e != cudaSuccess)
    {
        std::cout << "CUDA runtime API error " << cudaGetErrorName(e) << " at line " << iLine << " in file " << szFile << std::endl;
        return false;
    }
    return true;
}

int main()
{
    int in_width = 1024 * 20;
    int in_height = 1024;
    int itr_num = 1;
    int out_width = in_height;
    int out_height = in_width;
    float *input_data_host = nullptr;
    float *out_data_host = nullptr;
    float *input_data_dev = nullptr;
    float *output_data_dev = nullptr;
    cudaStream_t stream = nullptr;
    cudaEvent_t start, end;
    cudaStreamCreate(&stream);
    cudaEventCreate(&start);
    cudaEventCreate(&end);

    int input_size = in_width * in_height * sizeof(float);
    int output_size = out_width * out_height * sizeof(float);
    float *in_data_cpu = (float *)malloc(in_width * in_height * sizeof(float));
    float *out_data_cpu = (float *)malloc(out_width * out_height * sizeof(float));
    ck(cudaMallocHost((void **)&input_data_host, input_size));
    ck(cudaMalloc((void **)&input_data_dev, input_size));
    ck(cudaMalloc((void **)&output_data_dev, output_size));
    out_data_host = (float *)malloc(output_size);
    for (int i = 0; i < in_width * in_height; i++) {
        // input_data_host[i] = __float2half(float(i % 64));
        in_data_cpu[i] = __float2half(static_cast <float> (rand()) / (static_cast <float> (RAND_MAX/10)));
        // in_data_cpu[i] = static_cast<float>(i%64) * 0.1;
        input_data_host[i] = __float2half(in_data_cpu[i]);
    }
    memset(out_data_cpu, 0, output_size);
    cudaMemcpyAsync(input_data_dev, input_data_host, input_size, cudaMemcpyHostToDevice, stream);
    cudaEventRecord(start, stream);
    for (int itr = 0; itr < itr_num; itr++){
        transpose(input_data_dev, output_data_dev, in_height, in_width, stream);
    }
    cudaEventRecord(end, stream);
    cudaEventSynchronize(end);
    float elapsedTime = 0.0;
    cudaEventElapsedTime(&elapsedTime, start, end);
    printf("size:[%d,%d] elapsed time is %fms\n", in_height, in_width, elapsedTime/itr_num);
    cudaMemcpy(out_data_cpu, output_data_dev, output_size, cudaMemcpyDeviceToHost);
    for (int idx = 0; idx < out_height * out_width; ++idx) {
        int height = idx / out_width;
        int width = idx % out_width;
        float diff = abs(out_data_cpu[idx] - in_data_cpu[width * in_width + height]);
        if (diff > 1e-6) {
            printf("gpu rets[%d][%d] is %f cpu rest is %f diff is %f\n", height, width, out_data_cpu[idx], in_data_cpu[width * in_width + height], diff);
            printf("faild!\n");
            break;
        }
    }
    printf("Success!\n");

    ck(cudaFreeHost(input_data_host));
    free(out_data_host);
    ck(cudaFree(input_data_dev));
    ck(cudaFree(output_data_dev));
    ck(cudaStreamDestroy(stream));
    return 0;
}
