#include "transpose.h"
#include <cuda_runtime_api.h>
#include <cuda_runtime.h>

constexpr int offset = 0;
template<int TX, int TY>
__global__ void transpose_kernel(float *in, float *out, int in_height, int in_width)
{
    __shared__ float shmm[TY * (TX+offset)]; // 512 * 4 = 2048 2k
    int row = blockIdx.y * blockDim.y + threadIdx.y;
    int col = blockIdx.x * blockDim.x + threadIdx.x;
    int tx = threadIdx.y * blockDim.x + threadIdx.x;
    if (row < in_height & col < in_width) {
        shmm[threadIdx.y * (TX + offset) + threadIdx.x] = in[row * in_width + col];
    }
    __syncthreads();

    int row_i = tx / blockDim.y;
    int col_i = tx % blockDim.y;
    col = blockIdx.y * blockDim.y + col_i;
    row = blockIdx.x * blockDim.x + row_i;
    if (row < in_width & col < in_height) {
        out[row * in_height + col] = shmm[col_i * (TX + offset) + row_i];
        // if (tx / 32 != 0 && tx % 32 == (col_i * (TX + offset) + row_i)%32)
            // printf("tx[%d] (%d %d) bank %d\n",tx, tx / 32, tx % 32, (col_i * (TX + offset) + row_i)%32);
    }
}

void transpose(float *in_data, float *out_data, int in_height, int in_width, cudaStream_t stream)
{
    constexpr int TX = 32;
    constexpr int TY = 16;
    int GX = (in_width + TX - 1) / TX;
    int GY = (in_height + TY - 1) / TY;
    dim3 block(TX, TY, 1);
    dim3 grid(GX, GY, 1);
    transpose_kernel<TX, TY><<<grid, block, 0, stream>>>(in_data, out_data, in_height, in_width);
}

